#!/usr/bin/env python
# --!-- coding: utf8 --!--
from PyQt5.QtCore import Qt, QRect, QEvent, QCoreApplication
from PyQt5.QtGui import QPalette, QFontMetrics, QKeySequence
from PyQt5.QtWidgets import QWidget, qApp, QListWidgetItem, QStyledItemDelegate, QStyle, QLabel, QToolTip, QShortcut


from manuskript.functions import mainWindow
from manuskript.ui import style
from manuskript.ui.search_ui import Ui_search
from manuskript.enums import Model

from manuskript.models.flatDataModelWrapper import flatDataModelWrapper
from manuskript.ui.searchMenu import searchMenu
from manuskript.ui.highlighters.searchResultHighlighters.searchResultHighlighter import searchResultHighlighter


class search(QWidget, Ui_search):
    def __init__(self, parent=None):
        _translate = QCoreApplication.translate

        QWidget.__init__(self, parent)
        self.setupUi(self)

        self.searchTextInput.returnPressed.connect(self.search)

        self.searchMenu = searchMenu()
        self.btnOptions.setMenu(self.searchMenu)

        self.delegate = listResultDelegate(self)
        self.result.setItemDelegate(self.delegate)
        self.result.setMouseTracking(True)
        self.result.itemClicked.connect(self.openItem)

        self.result.setStyleSheet(style.searchResultSS())
        self.searchTextInput.setStyleSheet(style.lineEditSS())

        self.searchResultHighlighter = searchResultHighlighter()

        self.noResultsLabel = QLabel(_translate("Search", "No results found"), self.result)
        self.noResultsLabel.setVisible(False)
        self.noResultsLabel.setStyleSheet("QLabel {color: gray;}")

        # Add shortcuts for navigating through search results
        QShortcut(QKeySequence(_translate("MainWindow", "F3")), self.searchTextInput, self.nextSearchResult)
        QShortcut(QKeySequence(_translate("MainWindow", "Shift+F3")), self.searchTextInput, self.previousSearchResult)

        # These texts are already included in translation files but including ":" at the end. We force here the
        # translation for them without ":"
        _translate("MainWindow", "Situation")
        _translate("MainWindow", "Status")

    def nextSearchResult(self):
        if self.result.currentRow() < self.result.count() - 1:
            self.result.setCurrentRow(self.result.currentRow() + 1)
        else:
            self.result.setCurrentRow(0)

        if 0 < self.result.currentRow() < self.result.count():
            self.openItem(self.result.currentItem())

    def previousSearchResult(self):
        if self.result.currentRow() > 0:
            self.result.setCurrentRow(self.result.currentRow() - 1)
        else:
            self.result.setCurrentRow(self.result.count() - 1)

        if 0 < self.result.currentRow() < self.result.count():
            self.openItem(self.result.currentItem())

    def prepareRegex(self, searchText):
        import re

        flags = re.UNICODE

        if self.searchMenu.caseSensitive() is False:
            flags |= re.IGNORECASE

        if self.searchMenu.regex() is False:
            searchText = re.escape(searchText)

        if self.searchMenu.matchWords() is True:
            # Source: https://stackoverflow.com/a/15863102
            searchText = r'\b' + searchText + r'\b'

        return re.compile(searchText, flags)

    def search(self):
        self.result.clear()
        self.result.setCurrentRow(0)

        searchText = self.searchTextInput.text()
        if len(searchText) > 0:
            searchRegex = self.prepareRegex(searchText)
            results = []

            # Set override cursor
            qApp.setOverrideCursor(Qt.WaitCursor)

            for model, modelName in [
                (mainWindow().mdlOutline, Model.Outline),
                (mainWindow().mdlCharacter, Model.Character),
                (flatDataModelWrapper(mainWindow().mdlFlatData), Model.FlatData),
                (mainWindow().mdlWorld, Model.World),
                (mainWindow().mdlPlots, Model.Plot)
            ]:
                filteredColumns = self.searchMenu.columns(modelName)

                # Searching
                if len(filteredColumns):
                    results += model.searchOccurrences(searchRegex, filteredColumns)

            # Showing results
            self.generateResultsLists(results)

            # Remove override cursor
            qApp.restoreOverrideCursor()

    def generateResultsLists(self, results):
        self.noResultsLabel.setVisible(len(results) == 0)
        for result in results:
            item = QListWidgetItem(result.title(), self.result)
            item.setData(Qt.UserRole, result)
            item.setData(Qt.UserRole + 1, ' > '.join(result.path()))
            item.setData(Qt.UserRole + 2, result.context())
            self.result.addItem(item)

    def openItem(self, item):
        self.searchResultHighlighter.highlightSearchResult(item.data(Qt.UserRole))

    def leaveEvent(self, event):
        self.delegate.mouseLeave()

class listResultDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)
        self._tooltipRowIndex = -1

    def paint(self, painter, option, index):
        extra = index.data(Qt.UserRole + 1)

        if not extra:
            return QStyledItemDelegate.paint(self, painter, option, index)
        else:
            if option.state & QStyle.State_Selected:
                painter.fillRect(option.rect, option.palette.color(QPalette.Highlight))

            title = index.data()
            painter.drawText(option.rect.adjusted(2, 1, 0, 0), Qt.AlignLeft, title)

            fm = QFontMetrics(option.font)
            w = fm.width(title)
            r = QRect(option.rect)
            r.setLeft(r.left() + w)
            painter.save()
            if option.state & QStyle.State_Selected:
                painter.setPen(Qt.white)
            else:
                painter.setPen(Qt.gray)
            painter.drawText(r.adjusted(2, 1, 0, 0), Qt.AlignLeft, " - {}".format(extra))
            painter.restore()

    def editorEvent(self, event, model, option, index):
        if event.type() == QEvent.MouseMove and self._tooltipRowIndex != index.row():
            self._tooltipRowIndex = index.row()
            context = index.data(Qt.UserRole + 2)
            extra = index.data(Qt.UserRole + 1)
            QToolTip.showText(event.globalPos(),
                              "<p>#" + str(index.row()) + " - " + extra + "</p><p>" + context + "</p>")
            return True
        return False

    def mouseLeave(self):
        self._tooltipRowIndex = -1



#!/usr/bin/env python
# --!-- coding: utf8 --!--


from manuskript.ui.highlighters.searchResultHighlighters.widgetSelectionHighlighter import widgetSelectionHighlighter


class abstractSearchResultHighlighter():
    def __init__(self):
        self._widgetSelectionHighlighter = widgetSelectionHighlighter()

    def highlightSearchResult(self, searchResult):
        self.openView(searchResult)
        widgets = self.retrieveWidget(searchResult)
        if not isinstance(widgets, list):
            widgets = [widgets]
        for i in range(len(widgets)):
            self._widgetSelectionHighlighter.highlight_widget_selection(widgets[i], searchResult.pos()[i][0], searchResult.pos()[i][1], i == len(widgets) - 1)

    def openView(self, searchResult):
        raise RuntimeError

    def retrieveWidget(self, searchResult):
        raise RuntimeError



#!/usr/bin/env python
# --!-- coding: utf8 --!--

from manuskript.models import references as Ref
from manuskript.enums import Outline
from manuskript.ui.highlighters.searchResultHighlighters.abstractSpecificSearchResultHighlighter import abstractSearchResultHighlighter
from manuskript.functions import mainWindow
from PyQt5.QtWidgets import QTextEdit, QLineEdit, QLabel
from manuskript.ui.views.metadataView import metadataView
from manuskript.ui.collapsibleGroupBox2 import collapsibleGroupBox2


class outlineSearchResultHighlighter(abstractSearchResultHighlighter):
    def __init__(self):
        super().__init__()
        self.outline_index = None

    def openView(self, searchResult):
        r = Ref.textReference(searchResult.id())
        Ref.open(r)

    def retrieveWidget(self, searchResult):
        editors = {
            Outline.text: ("txtRedacText", QTextEdit, None),
            Outline.title: ("txtTitle", QLineEdit, "grpProperties"),
            Outline.summarySentence: ("txtSummarySentence", QLineEdit, "grpSummary"),
            Outline.summaryFull: ("txtSummaryFull", QTextEdit, "grpSummary"),
            Outline.notes: ("txtNotes", QTextEdit, "grpNotes"),

            # TODO: Tried to highlight the combo box themselves (ie. cmbPOV) but didn't succeed.
            Outline.POV: ("lblPOV", QLabel, "grpProperties"),
            Outline.status: ("lblStatus", QLabel, "grpProperties"),
            Outline.label: ("lblLabel", QLabel, "grpProperties")
        }

        editorName, editorClass, parentName = editors[searchResult.column()]

        # Metadata columns are inside a splitter widget that my be hidden, so we show them.
        if parentName:
            metadataViewWidget = mainWindow().findChild(metadataView, "redacMetadata")
            metadataViewWidget.show()
            metadataViewWidget.findChild(collapsibleGroupBox2, parentName).button.setChecked(True)
            widget = metadataViewWidget.findChild(editorClass, editorName)
        else:
            widget = mainWindow().mainEditor.currentEditor().findChild(editorClass, editorName)

        return widget



#!/usr/bin/env python
# --!-- coding: utf8 --!--

import pytest
from manuskript.models.searchFilter import searchFilter


def test_searchFilter_constructionOk():
    filter = searchFilter("label", True, [3])
    assert filter.label() == "label"
    assert filter.enabled() is True
    assert filter.modelColumns() == [3]


def test_searchFilter_constructionOkWithNoneModelColumn():
    filter = searchFilter("label", True)
    assert filter.label() == "label"
    assert filter.enabled() is True
    assert filter.modelColumns() == []


def test_searchFilter_constructionBadLabelType():
    with pytest.raises(TypeError, match=r".*label must be a str.*"):
        searchFilter(13, True, [3])


def test_searchFilter_constructionBadEnabledType():
    with pytest.raises(TypeError, match=r".*enabled must be a bool.*"):
        searchFilter("label", 3, [3])


def test_searchFilter_constructionBadModelColumnType():
    with pytest.raises(TypeError, match=r".*modelColumns must be a list or None.*"):
        searchFilter("label", False, True)


def test_searchFilter_setEnabled():
    filter = searchFilter("label", True, [3])
    assert filter.enabled() is True
    filter.setEnabled(False)
    assert filter.enabled() is False



#!/usr/bin/env python
# --!-- coding: utf8 --!--
from PyQt5.QtCore import QModelIndex
from PyQt5.QtCore import QSignalMapper
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QBrush
from PyQt5.QtGui import QStandardItem
from PyQt5.QtGui import QStandardItemModel
from PyQt5.QtWidgets import QAction, QMenu

from manuskript.enums import Plot, PlotStep, Model
from manuskript.functions import toInt, mainWindow
from manuskript.models.searchResultModel import searchResultModel
from manuskript.searchLabels import PlotSearchLabels, PLOT_STEP_COLUMNS_OFFSET
from manuskript.functions import search
from manuskript.models.searchableModel import searchableModel
from manuskript.models.searchableItem import searchableItem

class plotModel(QStandardItemModel, searchableModel):
    def __init__(self, parent):
        QStandardItemModel.__init__(self, 0, 3, parent)
        self.setHorizontalHeaderLabels([i.name for i in Plot])
        self.mw = mainWindow()

        self.updatePlotPersoButton()

    ###############################################################################
    # QUERIES
    ###############################################################################

    def getPlotsByImportance(self):
        plots = [[], [], []]
        for i in range(self.rowCount()):
            importance = self.item(i, Plot.importance).text()
            ID = self.item(i, Plot.ID).text()
            plots[2 - toInt(importance)].append(ID)
        return plots

    def getSubPlotsByID(self, ID):
        index = self.getIndexFromID(ID)
        if not index.isValid():
            return
        index = index.sibling(index.row(), Plot.steps)
        item = self.itemFromIndex(index)
        lst = []
        for i in range(item.rowCount()):
            if item.child(i, PlotStep.ID):
                _ID = item.child(i, PlotStep.ID).text()

                # Don't know why sometimes name is None (while drag'n'dropping
                # several items)
                if item.child(i, PlotStep.name):
                    name = item.child(i, PlotStep.name).text()
                else:
                    name = ""

                # Don't know why sometimes summary is None
                if item.child(i, PlotStep.summary):
                    summary = item.child(i, PlotStep.summary).text()
                else:
                    summary = ""

                lst.append((_ID, name, summary))
        return lst

    def getPlotNameByID(self, ID):
        for i in range(self.rowCount()):
            _ID = self.item(i, Plot.ID).text()
            if _ID == ID or toInt(_ID) == ID:
                name = self.item(i, Plot.name).text()
                return name
        return None

    def getPlotImportanceByRow(self, row):
        for i in range(self.rowCount()):
            if i == row:
                importance = self.item(i, Plot.importance).text()
                return importance
        return "0"  # Default to "Minor"

    def getSubPlotTextsByID(self, plotID, subplotRaw):
        """Returns a tuple (name, summary) for the subplot whose raw in the model
        is ``subplotRaw``, of plot whose ID is ``plotID``.
        """
        plotIndex = self.getIndexFromID(plotID)
        name = plotIndex.child(subplotRaw, PlotStep.name).data()
        summary = plotIndex.child(subplotRaw, PlotStep.summary).data()
        return name, summary

    def getIndexFromID(self, ID):
        for i in range(self.rowCount()):
            _ID = self.item(i, Plot.ID).text()
            if _ID == ID or toInt(_ID) == ID:
                return self.index(i, 0)
        return QModelIndex()

    def currentIndex(self):
        i = self.mw.lstPlots.currentIndex()
        if i.isValid():
            return i
        else:
            return None

    ###############################################################################
    # ADDING / REMOVING
    ###############################################################################

    def addPlot(self, name="New plot"):
        if not name:
            name="New Plot"
        p = QStandardItem(self.tr(name))
        _id = QStandardItem(self.getUniqueID())
        importance = QStandardItem(str(0))
        self.appendRow([p, _id, importance, QStandardItem("Characters"),
                        QStandardItem(), QStandardItem(), QStandardItem("Resolution steps")])
        return p, _id

    def getUniqueID(self, parent=QModelIndex()):
        """Returns an unused ID"""
        parentItem = self.itemFromIndex(parent)
        vals = []
        for i in range(self.rowCount(parent)):
            index = self.index(i, Plot.ID, parent)
            # item = self.item(i, Plot.ID)
            if index.isValid() and index.data():
                vals.append(int(index.data()))

        k = 0
        while k in vals:
            k += 1
        return str(k)

    def removePlot(self, index):
        self.takeRow(index.row())

    ###############################################################################
    # SUBPLOTS
    ###############################################################################

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            if orientation == Qt.Horizontal:
                if section == PlotStep.name:
                    return self.tr("Name")
                elif section == PlotStep.meta:
                    return self.tr("Meta")
                else:
                    return ""
            else:
                return ""
        else:
            return QStandardItemModel.headerData(self, section, orientation, role)

    def data(self, index, role=Qt.DisplayRole):
        if index.parent().isValid() and \
                index.parent().column() == Plot.steps and \
                index.column() == PlotStep.meta:
            if role == Qt.TextAlignmentRole:
                return Qt.AlignRight | Qt.AlignVCenter
            elif role == Qt.ForegroundRole:
                return QBrush(Qt.gray)
            else:
                return QStandardItemModel.data(self, index, role)

        else:
            return QStandardItemModel.data(self, index, role)

    def addSubPlot(self):
        index = self.mw.lstPlots.currentPlotIndex()
        if not index.isValid():
            return

        parent = index.sibling(index.row(), Plot.steps)
        parentItem = self.item(index.row(), Plot.steps)

        if not parentItem:
            return

        p = QStandardItem(self.tr("New step"))
        _id = QStandardItem(self.getUniqueID(parent))
        summary = QStandardItem()

        currentIndex = self.mw.lstSubPlots.selectionModel().selectedIndexes()
        if currentIndex:
            # We use last item of selection in case of many
            currentIndex = currentIndex[-1]
            row = currentIndex.row() + 1
            parentItem.insertRow(row, [p, _id, QStandardItem(), summary])
            # Select last index
            self.mw.lstSubPlots.setCurrentIndex(currentIndex.sibling(row, 0))
        else:
            # Don't know why, if summary is in third position, then drag/drop deletes it...
            parentItem.appendRow([p, _id, QStandardItem(), summary])
            # Select last index
            self.mw.lstSubPlots.setCurrentIndex(
                parent.child(self.rowCount(parent) - 1, 0))

    def removeSubPlot(self):
        """
        Remove all selected subplots / plot steps, in mw.lstSubPlots.
        """
        parent = self.mw.lstSubPlots.rootIndex()
        if not parent.isValid():
            return
        parentItem = self.itemFromIndex(parent)

        while self.mw.lstSubPlots.selectionModel().selectedRows():
            i = self.mw.lstSubPlots.selectionModel().selectedRows()[0]
            parentItem.takeRow(i.row())

    def flags(self, index):
        parent = index.parent()
        if parent.isValid():  # this is a subitem
            return Qt.ItemIsEnabled | Qt.ItemIsEditable | Qt.ItemIsSelectable | Qt.ItemIsDragEnabled
        else:
            return QStandardItemModel.flags(self, index)

    ###############################################################################
    # PLOT PERSOS
    ###############################################################################

    def addPlotPerso(self, v):
        index = self.mw.lstPlots.currentPlotIndex()
        if index.isValid():
            if not self.item(index.row(), Plot.characters):
                self.setItem(index.row(), Plot.characters, QStandardItem())

            item = self.item(index.row(), Plot.characters)

            # We check that the PersoID is not in the list yet
            for i in range(item.rowCount()):
                if item.child(i).text() == str(v):
                    return

            item.appendRow(QStandardItem(str(v)))

    def removePlotPerso(self):
        index = self.mw.lstPlotPerso.currentIndex()
        if not index.isValid():
            return
        parent = index.parent()
        parentItem = self.itemFromIndex(parent)
        parentItem.takeRow(index.row())

    def updatePlotPersoButton(self):
        menu = QMenu(self.mw)

        menus = []
        for i in [self.tr("Main"), self.tr("Secondary"), self.tr("Minor")]:
            m = QMenu(i, menu)
            menus.append(m)
            menu.addMenu(m)

        mpr = QSignalMapper(menu)
        for i in range(self.mw.mdlCharacter.rowCount()):
            a = QAction(self.mw.mdlCharacter.name(i), menu)
            a.setIcon(self.mw.mdlCharacter.icon(i))
            a.triggered.connect(mpr.map)
            mpr.setMapping(a, int(self.mw.mdlCharacter.ID(i)))

            imp = toInt(self.mw.mdlCharacter.importance(i))

            menus[2 - imp].addAction(a)

        # Disabling empty menus
        for m in menus:
            if not m.actions():
                m.setEnabled(False)

        mpr.mapped.connect(self.addPlotPerso)
        self.mw.btnAddPlotPerso.setMenu(menu)

    #######################################################################
    # Search
    #######################################################################
    def searchableItems(self):
        items = []

        for i in range(self.rowCount()):
            items.append(plotItemSearchWrapper(i, self.item, self.mw.mdlCharacter.getCharacterByID))

        return items


class plotItemSearchWrapper(searchableItem):
    def __init__(self, rowIndex, getItem, getCharacterByID):
        self.rowIndex = rowIndex
        self.getItem = getItem
        self.getCharacterByID = getCharacterByID
        super().__init__(PlotSearchLabels)

    def searchOccurrences(self, searchRegex, column):
        results = []

        plotName = self.getItem(self.rowIndex, Plot.name).text()
        if column >= PLOT_STEP_COLUMNS_OFFSET:
            results += self.searchInPlotSteps(self.rowIndex, plotName, column, column - PLOT_STEP_COLUMNS_OFFSET, searchRegex, False)
        else:
            item_name = self.getItem(self.rowIndex, Plot.name).text()
            if column == Plot.characters:
                charactersList = self.getItem(self.rowIndex, Plot.characters)

                for i in range(charactersList.rowCount()):
                    characterID = charactersList.child(i).text()

                    character = self.getCharacterByID(characterID)
                    if character:
                        columnText = character.name()

                        characterResults = search(searchRegex, columnText)
                        if len(characterResults):
                            # We will highlight the full character row in the plot characters list, so we
                            # return the row index instead of the match start and end positions.
                            results += [
                                searchResultModel(Model.Plot, self.getItem(self.rowIndex, Plot.ID).text(), column,
                                             self.translate(item_name),
                                             self.searchPath(column),
                                             [(i, 0)], context) for start, end, context in
                                search(searchRegex, columnText)]
            else:
                results += super().searchOccurrences(searchRegex, column)
                if column == Plot.name:
                    results += self.searchInPlotSteps(self.rowIndex, plotName, Plot.name, PlotStep.name,
                                                      searchRegex, False)
                elif column == Plot.summary:
                    results += self.searchInPlotSteps(self.rowIndex, plotName, Plot.summary, PlotStep.summary,
                                                      searchRegex, True)

        return results

    def searchModel(self):
        return Model.Plot

    def searchID(self):
        return self.getItem(self.rowIndex, Plot.ID).text()

    def searchTitle(self, column):
        return self.getItem(self.rowIndex, Plot.name).text()

    def searchPath(self, column):
        def _path(item):
            path = []

            if item.parent():
                path += _path(item.parent())
            path.append(item.text())

            return path

        return [self.translate("Plot")] + _path(self.getItem(self.rowIndex, Plot.name)) + [self.translate(self.searchColumnLabel(column))]

    def searchData(self, column):
        return self.getItem(self.rowIndex, column).text()

    def plotStepPath(self, plotName, plotStepName, column):
        return [self.translate("Plot"), plotName, plotStepName, self.translate(self.searchColumnLabel(column))]

    def searchInPlotSteps(self, plotIndex, plotName, plotColumn, plotStepColumn, searchRegex, searchInsidePlotStep):
        results = []

        # Plot step info can be found in two places: the own list of plot steps (this is the case for ie. name and meta
        # fields) and "inside" the plot step once it is selected in the list (as it's the case for the summary).
        if searchInsidePlotStep:
            # We are searching *inside* the plot step, so we return both the row index (for selecting the right plot
            # step in the list), and (start, end) positions of the match inside the text field for highlighting it.
            getSearchData = lambda rowIndex, start, end, context: ([(rowIndex, 0), (start, end)], context)
        else:
            # We are searching *in the plot step row*, so we only return the row index for selecting the right plot
            # step in the list when highlighting search results.
            getSearchData = lambda rowIndex, start, end, context: ([(rowIndex, 0)], context)

        item = self.getItem(plotIndex, Plot.steps)
        for i in range(item.rowCount()):
            if item.child(i, PlotStep.ID):
                plotStepName = item.child(i, PlotStep.name).text()
                plotStepText = item.child(i, plotStepColumn).text()

                # We will highlight the full plot step row in the plot steps list, so we
                # return the row index instead of the match start and end positions.
                results += [searchResultModel(Model.PlotStep, self.getItem(plotIndex, Plot.ID).text(), plotStepColumn,
                                         self.translate(plotStepName),
                                         self.plotStepPath(plotName, plotStepName, plotColumn),
                                         *getSearchData(i, start, end, context)) for start, end, context in
                            search(searchRegex, plotStepText)]

        return results


#!/usr/bin/env python
# --!-- coding: utf8 --!--


from manuskript.models import references as Ref
from manuskript.functions import mainWindow
from manuskript.enums import Character
from PyQt5.QtWidgets import QTextEdit, QTableView, QLineEdit
from manuskript.ui.highlighters.searchResultHighlighters.abstractSpecificSearchResultHighlighter import abstractSearchResultHighlighter


class characterSearchResultHighlighter(abstractSearchResultHighlighter):
    def __init__(self):
        super().__init__()

    def openView(self, searchResult):
        r = Ref.characterReference(searchResult.id())
        Ref.open(r)
        mainWindow().tabPersos.setEnabled(True)

    def retrieveWidget(self, searchResult):
        textEditMap = {
            Character.name: (0, "txtPersoName", QLineEdit),
            Character.goal: (0, "txtPersoGoal", QTextEdit),
            Character.motivation: (0, "txtPersoMotivation", QTextEdit),
            Character.conflict: (0, "txtPersoConflict", QTextEdit),
            Character.epiphany: (0, "txtPersoEpiphany", QTextEdit),
            Character.summarySentence: (0, "txtPersoSummarySentence", QTextEdit),
            Character.summaryPara: (0, "txtPersoSummaryPara", QTextEdit),
            Character.summaryFull: (1, "txtPersoSummaryFull", QTextEdit),
            Character.notes: (2, "txtPersoNotes", QTextEdit),
            Character.infos: (3, "tblPersoInfos", QTableView)
        }

        characterTabIndex, characterWidgetName, characterWidgetClass = textEditMap[searchResult.column()]

        mainWindow().tabPersos.setCurrentIndex(characterTabIndex)
        return mainWindow().tabPersos.findChild(characterWidgetClass, characterWidgetName)



#!/usr/bin/env python
# --!-- coding: utf8 --!--


class searchableModel():

    def searchOccurrences(self, searchRegex, columns):
        results = []
        for item in self.searchableItems():
            for column in columns:
                results += item.searchOccurrences(searchRegex, column)
        return results

    def searchableItems(self):
        raise NotImplementedError



#!/usr/bin/env python
# --!-- coding: utf8 --!--

from manuskript.ui.searchMenu import searchMenu
from manuskript.enums import Outline, Character, FlatData, World, Plot, PlotStep, Model
from manuskript.searchLabels import PLOT_STEP_COLUMNS_OFFSET


def triggerFilter(filterKey, actions):
    list(filter(lambda action: action.data() == filterKey, actions))[0].trigger()


def test_searchMenu_defaultColumns():
    """
    By default all model columns are selected.
    """
    search_menu = searchMenu()

    assert set(search_menu.columns(Model.Outline)) == {
        Outline.title, Outline.text, Outline.summaryFull,
        Outline.summarySentence, Outline.notes, Outline.POV,
        Outline.status, Outline.label
    }

    assert set(search_menu.columns(Model.Character)) == {
        Character.name, Character.motivation, Character.goal, Character.conflict,
        Character.epiphany, Character.summarySentence, Character.summaryPara,
        Character.summaryFull, Character.notes, Character.infos
    }

    assert set(search_menu.columns(Model.FlatData)) == {
        FlatData.summarySituation, FlatData.summarySentence, FlatData.summaryPara,
        FlatData.summaryPage, FlatData.summaryFull
    }

    assert set(search_menu.columns(Model.World)) == {
        World.name, World.description, World.passion, World.conflict
    }

    assert set(search_menu.columns(Model.Plot)) == {
        Plot.name, Plot.description, Plot.characters, Plot.result,
        Plot.summary, PLOT_STEP_COLUMNS_OFFSET + PlotStep.meta
    }


def test_searchMenu_someColumns():
    """
    When deselecting some filters the columns associated to those filters are not returned.
    """
    search_menu = searchMenu()

    triggerFilter(Model.Outline, search_menu.actions())
    triggerFilter(Model.Character, search_menu.actions())

    assert set(search_menu.columns(Model.Outline)) == set()
    assert set(search_menu.columns(Model.Character)) == set()



# -*- coding: utf-8 -*-

# While all logging should be done through the facilities offered by the
# standard python `logging` module, this module will take care of specific
# manuskript needs to keep it separate from the rest of the logic.

import os
import sys
import time
import logging
import pathlib

from manuskript.functions import writablePath
from importlib import  import_module
from pprint import pformat

LOGGER = logging.getLogger(__name__)

LOGFORMAT_CONSOLE = "%(levelname)s> %(message)s"
LOGFORMAT_FILE = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"

def setUp(console_level=logging.WARN):
    """Sets up a convenient environment for logging.

    To console:  >WARNING, plain.        (Only the essence.)"""

    # The root_logger should merely trigger on warnings since it is the final
    # stop after all categories we really care about didn't match.
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.WARN)
    # The manuskript_logger is what all of our own code will come by.
    # Obviously, we care greatly about logging every single message.
    manuskript_logger = logging.getLogger("manuskript")
    manuskript_logger.setLevel(logging.DEBUG)
    # The qt_logger sees all the Qt nonsense when it breaks.
    # We don't really want to know... but we have to know.
    qt_logger = logging.getLogger("qt")
    qt_logger.setLevel(logging.DEBUG)

    # Send logs of WARNING+ to STDERR for higher visibility.
    ch = logging.StreamHandler()
    ch.setLevel(console_level)
    ch.setFormatter(logging.Formatter(LOGFORMAT_CONSOLE))
    root_logger.addHandler(ch)

    # Any exceptions we did not account for need to be logged.
    logFutureExceptions()

    LOGGER.debug("Logging to STDERR.")


def getDefaultLogFile():
    """Returns a filename to log to inside {datadir}/logs/.

    It also prunes old logs so that we do not hog disk space excessively over time.
    """
    # Ensure logs directory exists.
    logsPath = os.path.join(writablePath(), "logs")
    os.makedirs(logsPath, exist_ok=True)
    # Prune irrelevant log files. They are only kept for 35 days.
    try:  # Guard against os.scandir() in the name of paranoia.
        now = time.time()
        with os.scandir(logsPath) as it:
            for f in it:
                try:  # Avoid triggering outer try-except inside loop.
                    if f.is_dir():
                        continue  # If a subdirectory exists for whatever reason, don't touch it.
                    if (now - f.stat().st_ctime) // (24 * 3600) >= 35:
                        os.remove(f)
                except OSError:
                    continue  # Fail silently, but make sure we check other files.
    except OSError:
        pass  # Fail silently. Don't explode and prevent Manuskript from starting.
    return os.path.join(logsPath, "%Y-%m-%d_%H-%M-%S_manuskript#%#.log")


def formatLogName(formatString, pid=None, now=None):
    """A minor hack on top of `strftime()` to support an identifier for the process ID.

    We want to support this in case some genius manages to start two manuskript processes
    during the exact same second, causing a conflict in log filenames.

    Additionally, there is a tiny chance that the pid could actually end up relevant when
    observing strange behaviour with a Manuskript process but having multiple instances open.
    """
    if pid == None:
        pid = os.getpid()
    if now == None:
        now = time.localtime()

    # Replace %# that is NOT preceded by %. Although this is not a perfect solution,
    # it is good enough because it is unlikely anyone would want to format '%pid'.
    lidx = 0
    while True:  # This could be neater with the := operator of Python 3.8 ...
        fidx = formatString.find("%#", lidx)
        if fidx == -1:
            break
        elif (fidx == 0) or (formatString[fidx-1] != "%"):
            formatString = formatString[:fidx] + str(pid) + formatString[fidx+2:]
            lidx = fidx + len(str(pid)) - 2
        else:  # skip and avoid endless loop
            lidx = fidx + 1

    # Finally apply strftime normally.
    return time.strftime(formatString, now)


def logToFile(file_level=logging.DEBUG, logfile=None):
    """Sets up the FileHandler that logs to a file.

    This is being done separately due to relying on QApplication being properly
    configured; without it we cannot detect the proper location for the log file.

    To log file: >DEBUG, timestamped.    (All the details.)"""

    if logfile is None:
        logfile = getDefaultLogFile()

    logfile = formatLogName(logfile)

    # Log with extreme prejudice; everything goes to the log file.
    # Because Qt gave me a megabyte-sized logfile while testing, it
    # makes sense that the default behaviour of appending to existing
    # log files may not be in our users best interest for the time
    # being. (Unfortunately.)
    try:
        fh = logging.FileHandler(logfile, mode='w', encoding='utf-8')
        fh.setLevel(file_level)
        fh.setFormatter(logging.Formatter(LOGFORMAT_FILE))

        root_logger = logging.getLogger()
        root_logger.addHandler(fh)

        # Use INFO level to make it easier to find for users.
        LOGGER.info("Logging to file: %s", logfile)
    except Exception as ex:
        LOGGER.warning("Cannot log to file '%s'. Reason: %s", logfile, ex)


def getLogFilePath():
    """Extracts a filename we are logging to from the first FileHandler we find."""
    root_logger = logging.getLogger()
    for handler in root_logger.handlers:
        if isinstance(handler, logging.FileHandler):
            return handler.baseFilename
    return None


# Log uncaught and unraisable exceptions.

# Uncaught exceptions trigger moments before a thread is terminated due to
# an uncaught exception. It is the final stop, and as such is very likely
# to be the reason Manuskript suddenly closed on the user without warning.
# (It can also happen on other threads, but it is a bad thing regardless!)
def handle_uncaught_exception(exc_type, exc_value, exc_traceback):
    # Allow Ctrl+C for script execution to keep functioning as-is.
    if issubclass(exc_type, KeyboardInterrupt):
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        return # default exception hook handled it

    # Anything that reaches this handler can be considered a deal-breaker.
    LOGGER.critical("An unhandled exception has occurred!", exc_info=(exc_type, exc_value, exc_traceback))

    # Exit the program to preserve PyQt 'functionality' that is broken by
    # having our own uncaught exception hook. For more information, see:
    #   https://stackoverflow.com/questions/49065371/why-does-sys-excepthook-behave-differently-when-wrapped
    sys.exit(1)

    # Note that without it, unhandled Python exceptions thrown while in the
    # bowels of Qt may be written to the log multiple times. Under the motto
    # of failing fast and not having a misleading log file, this appears to
    # be the best course of action.


# The situation with threads and uncaught exceptions is fraught in peril.
# Hopefully this solves our problems on more recent versions of Python.
def handle_uncaught_thread_exception(args):
    if issubclass(exc_type, SystemExit):
        return # match behaviour of default hook, see manual

    # Anything that reaches this handler can be considered a minor deal-breaker.
    LOGGER.error("An unhandled exception has occurred in a thread: %s", repr(args.thread),
                    exc_info=(args.exc_type, args.exc_value, args.exc_traceback))


# Unraisable exceptions are exceptions that failed to be raised to a caller
# due to the nature of the exception. Examples: __del__(), GC error, etc.
# Logging these may expose bugs / errors that would otherwise go unnoticed.
def handle_unraisable_exception(unraisable):
    # Log as warning because the application is likely to limp along with
    # no serious side effects; a resource leak is the most likely.
    LOGGER.warning("%s: %s", unraisable.err_msg or "Exception ignored in", repr(unraisable.object),
                    exc_info=(unraisable.exc_type, unraisable.exc_value, unraisable.exc_traceback))


# Because we are already somewhat careful in regards to the order of code
# execution when it comes to setting up the logging environment, this has
# been put in its own function as opposed to letting a direct import handle it.
def logFutureExceptions():
    """Log all the interesting exceptions that may happen in the future."""
    sys.excepthook = handle_uncaught_exception
    try:
        import threading # threading module was optional pre-3.7
        if hasattr(threading, "excepthook"):  # Python 3.8+
            threading.excepthook = handle_uncaught_thread_exception
    except:
        pass
    if hasattr(sys, "unraisablehook"):  # Python 3.8+
        sys.unraisablehook = handle_unraisable_exception


# Qt has its own logging facility that we would like to integrate into our own.
# See: http://thispageintentionally.blogspot.com/2014/03/trapping-qt-log-messages.html

from PyQt5.QtCore import qInstallMessageHandler, QLibraryInfo, QMessageLogContext
from PyQt5.Qt import QtMsgType

def qtMessageHandler(msg_type, msg_log_context, msg_string):
    """Forwards Qt messages to Python logging system."""
    # Convert Qt msg type to logging level
    log_level = [logging.DEBUG,
                 logging.WARNING,
                 logging.ERROR,
                 logging.FATAL] [ int(msg_type) ]
    qtcl = logging.getLogger(msg_log_context.category or "qt.???")
    # Some information may not be available unless using a PyQt debug build.
    # See: https://www.riverbankcomputing.com/static/Docs/PyQt5/api/qtcore/qmessagelogcontext.html
    if QLibraryInfo.isDebugBuild():
        qtcl.log(logging.DEBUG,
                    ' @ {0} : {1}'.format((msg_log_context.file or "<unknown source file>"), msg_log_context.line)
                    )
        qtcl.log(logging.DEBUG,
                    ' ! {0}'.format((msg_log_context.function or "<unknown function>"))
                    )
    qtcl.log(log_level, msg_string)

def integrateQtLogging():
    """Integrates Qt logging facilities to be a part of our own."""

    # Note: the qtlogger is initialized in setUp() because it fits in
    # nicely with the initialization of the other loggers over there.
    # I also feel a lot safer this way. Qt is a curse that just keeps
    # on giving, even when it isn't actually at fault. I hate you, Qt.

    qInstallMessageHandler(qtMessageHandler)


def versionTupleToString(t):
    """A bit of generic tuple conversion code that hopefully handles all the
    different sorts of tuples we may come across while logging versions.

    None                -> "N/A"
    (,)                 -> "N/A"
    (2, 4, 6)           -> "2.4.6"
    (2, 4, "alpha", 8)  -> "2.4-alpha.8"
    """

    s = []
    if t is None or len(t) == 0:
        return "N/A"
    else:
        s.append(str(t[0]))

    def version_chunk(v):
        if isinstance(v, str):
            return "-", str(v)
        else:
            return ".", str(v)

    s.extend(f for p in t[1:] for f in version_chunk(p))
    return "".join(s)

def attributesFromOptionalModule(module, *attributes):
    """It is nice to cut down on the try-except boilerplate by
    putting this logic into its own function.

    Returns as many values as there are attributes.
    A value will be None if it failed to get the attribute."""

    assert(len(attributes) != 0)
    v = []
    try:
        m = import_module(module)

        for a in attributes:
            v.append(getattr(m, a, None))
    except ImportError:
        v.extend(None for _ in range(len(attributes)))

    if len(v) == 1:
        # Return the value directly so we can use it in an expression.
        return v[0]
    else:
        # The list is consumed as a part of the unpacking syntax.
        return v

def logRuntimeInformation(logger=None):
    """Logs all important runtime information neatly together.

    Due to the generic nature, use the manuskript logger by default."""

    if not logger:
        logger = logging.getLogger("manuskript")

    vt2s = versionTupleToString
    afom = attributesFromOptionalModule

    # Basic system information.
    from platform import python_version, platform, processor, machine
    logger.info("Operating System: %s", platform())
    logger.info("Hardware: %s / %s", machine(), processor())

    # Information about the running instance. See:
    #   https://pyinstaller.readthedocs.io/en/v3.3.1/runtime-information.html
    #   http://www.py2exe.org/index.cgi/Py2exeEnvironment
    #   https://cx-freeze.readthedocs.io/en/latest/faq.html#data-files
    frozen = getattr(sys, 'frozen', False)
    if frozen:
        logger.info("Running in a frozen (packaged) state.")
        logger.debug("* sys.frozen = %s", pformat(frozen))

        # PyInstaller, py2exe and cx_Freeze modules are not accessible while frozen,
        # so logging their version is (to my knowledge) impossible without including
        # special steps into the distribution process. But some traces do exist...
        logger.debug("* sys._MEIPASS = %s", getattr(sys, '_MEIPASS', "N/A"))  # PyInstaller bundle
        # cx_Freeze and py2exe do not appear to leave anything similar exposed.
    else:
        logger.info("Running from unpackaged source code.")

    # File not found? These bits of information might help.
    logger.debug("* sys.executable = %s", pformat(sys.executable))
    logger.debug("* sys.argv = %s", pformat(sys.argv))
    logger.debug("* sys.path = %s", pformat(sys.path))
    logger.debug("* sys.prefix = %s", pformat(sys.prefix))

    # Manuskript and Python info.
    from manuskript.functions import getGitRevisionAsString, getManuskriptPath
    from manuskript.version import getVersion
    logger.info("Manuskript %s%s (Python %s)", getVersion(),
                    getGitRevisionAsString(getManuskriptPath(), short=True),
                    python_version())

    # Installed Python packages.

    # PyQt + Qt
    from PyQt5.Qt import PYQT_VERSION_STR, qVersion
    from PyQt5.QtCore import QT_VERSION_STR
    logger.info("* PyQt %s (compiled against Qt %s)", PYQT_VERSION_STR, QT_VERSION_STR)
    logger.info("  * Qt %s (runtime)", qVersion())

    # Lxml
    # See: https://lxml.de/FAQ.html#i-think-i-have-found-a-bug-in-lxml-what-should-i-do
    from lxml import etree
    logger.info("* lxml.etree %s",                vt2s(etree.LXML_VERSION))
    logger.info("  * libxml   %s (compiled: %s)", vt2s(etree.LIBXML_VERSION), vt2s(etree.LIBXML_COMPILED_VERSION))
    logger.info("  * libxslt  %s (compiled: %s)", vt2s(etree.LIBXSLT_VERSION), vt2s(etree.LIBXSLT_COMPILED_VERSION))

    # Spellcheckers. (Optional)
    enchant_mod_ver, enchant_lib_ver = afom("enchant", "__version__", "get_enchant_version")
    if enchant_lib_ver:
        enchant_lib_ver = enchant_lib_ver()
        if isinstance(enchant_lib_ver, bytes):  # PyEnchant version < 3.0.2
            enchant_lib_ver = enchant_lib_ver.decode('utf-8')
    logger.info("* pyEnchant %s (libenchant: %s)", enchant_mod_ver or "N/A", enchant_lib_ver or "N/A")

    logger.info("* pySpellChecker %s", afom("spellchecker", "__version__") or "N/A")
    logger.info("* Symspellpy %s", afom("symspellpy", "__version__") or "N/A")

    # Markdown. (Optional)
    logger.info("* Markdown %s", afom("markdown", "__version__") or "N/A")

    # Web rendering engine
    from manuskript.ui.views.webView import webEngine
    logger.info("Web rendering engine: %s", webEngine)

    # Do not collect version information for Pandoc; that would require
    # executing `pandov -v` and parsing the output, all of which is too slow.



#!/usr/bin/env python
# --!-- coding: utf8 --!--
from PyQt5.QtCore import QModelIndex, Qt, QAbstractItemModel, QVariant
from PyQt5.QtGui import QIcon, QPixmap, QColor

from manuskript.functions import randomColor, iconColor, mainWindow, search
from manuskript.enums import Character as C, Model
from manuskript.searchLabels import CharacterSearchLabels

from manuskript.models.searchableModel import searchableModel
from manuskript.models.searchableItem import searchableItem

class characterModel(QAbstractItemModel, searchableModel):

    def __init__(self, parent):
        QAbstractItemModel.__init__(self, parent)

        # CharacterItems are stored in this list
        self.characters = []

###############################################################################
# QAbstractItemModel subclassed
###############################################################################

    def rowCount(self, parent=QModelIndex()):
        if parent.isValid():
            c = parent.internalPointer()
            return len(c.infos)
        else:
            return len(self.characters)

    def columnCount(self, parent=QModelIndex()):
        if parent.isValid():
            # Returns characters infos
            return 2
        else:
            return len(C)

    def data(self, index, role=Qt.DisplayRole):
        c = index.internalPointer()
        if type(c) == Character:
            if role == Qt.DisplayRole:
                if index.column() in c._data:
                    return c._data[index.column()]
                else:
                    return ""

            elif role == Qt.DecorationRole:
                if index.column() == C.name.value:
                    return c.icon
                else:
                    return QVariant()

        elif type(c) == CharacterInfo:
            if role == Qt.DisplayRole or role == Qt.EditRole:
                if index.column() == 0:
                    return c.description
                elif index.column() == 1:
                    return c.value

    def setData(self, index, value, role=Qt.EditRole):
        c = index.internalPointer()
        if type(c) == Character:
            if role == Qt.EditRole:
                # We update only if data is different
                if index.column() not in c._data or c._data[index.column()] != value:
                    c._data[index.column()] = value
                    self.dataChanged.emit(index, index)
                    return True

        elif type(c) == CharacterInfo:
            if role == Qt.EditRole:
                if index.column() == 0:
                    c.description = value
                elif index.column() == 1:
                    c.value = value
                self.dataChanged.emit(index, index)
                return True

        return False

    def index(self, row, column, parent=QModelIndex()):
        if not parent.isValid():
            return self.createIndex(row, column, self.characters[row])

        else:
            c = parent.internalPointer()
            if row < len(c.infos):
                return self.createIndex(row, column, c.infos[row])
            else:
                return QModelIndex()

    def indexFromItem(self, item, column=0):
        if not item:
            return QModelIndex()

        row = self.characters.index(item)
        col = column
        return self.createIndex(row, col, item)

    def parent(self, index):
        if not index.isValid():
            return QModelIndex()

        child = index.internalPointer()

        if type(child) == Character:
            return QModelIndex()

        elif type(child) == CharacterInfo:
            return child.character.index()

    def flags(self, index):
        if index.parent().isValid():
            return QAbstractItemModel.flags(self, index) | Qt.ItemIsEditable
        else:
            return QAbstractItemModel.flags(self, index)

###############################################################################
# CHARACTER QUERIES
###############################################################################

    def character(self, row):
        return self.characters[row]

    def name(self, row):
        return self.character(row).name()

    def icon(self, row):
        return self.character(row).icon

    def ID(self, row):
        return self.character(row).ID()

    def importance(self, row):
        return self.character(row).importance()

    def pov(self, row):
        return self.character(row).pov()

###############################################################################
# MODEL QUERIES
###############################################################################

    def getCharactersByImportance(self):
        """
        Lists characters by importance.

        @return: array of array of ´character´, by importance.
        """
        r = [[], [], []]

        for c in self.characters:
            r[2-int(c.importance())].append(c)

        return r

    def getCharacterByID(self, ID):
        if ID != None:
            ID = str(ID)
            for c in self.characters:
                if c.ID() == ID:
                    return c

        return None

###############################################################################
# ADDING / REMOVING
###############################################################################

    def addCharacter(self, importance = 0, name="New character"):
        """
        Creates a new character
        @param importance: the importance level of the character
        @return: the character
        """
        if not name:
            name="New Character"
        c = Character(model=self, name=self.tr(name), importance = importance)
        self.beginInsertRows(QModelIndex(), len(
            self.characters), len(self.characters))
        self.characters.append(c)
        self.endInsertRows()
        return c

    def removeCharacter(self, ID):
        """
        Removes character whose ID is ID...
        @param ID: the ID of the character to remove
        @return: nothing
        """
        c = self.getCharacterByID(ID)
        self.beginRemoveRows(QModelIndex(), self.characters.index(
            c), self.characters.index(c))
        self.characters.remove(c)
        self.endRemoveRows()

###############################################################################
# CHARACTER INFOS
###############################################################################

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            if section == 0:
                return self.tr("Name")
            elif section == 1:
                return self.tr("Value")
            else:
                return C(section).name

    def addCharacterInfo(self, ID):
        c = self.getCharacterByID(ID)
        self.beginInsertRows(c.index(), len(c.infos), len(c.infos))
        c.infos.append(CharacterInfo(
            c, description="Description", value="Value"))
        self.endInsertRows()

        mainWindow().updatePersoInfoView()

    def removeCharacterInfo(self, ID):
        c = self.getCharacterByID(ID)

        rm = []
        for idx in mainWindow().tblPersoInfos.selectedIndexes():
            if not idx.row() in rm:
                rm.append(idx.row())

        rm.sort()
        rm.reverse()
        for r in rm:
            self.beginRemoveRows(c.index(), r, r)
            c.infos.pop(r)
            self.endRemoveRows()

    def searchableItems(self):
        return self.characters

###############################################################################
# CHARACTER
###############################################################################

class Character(searchableItem):
    def __init__(self, model, name="No name", importance = 0):
        self._model = model
        self.lastPath = ""

        self._data = {}
        self._data[C.name.value] = name
        self.assignUniqueID()
        self.assignRandomColor()
        self._data[C.importance.value] = str(importance)
        self._data[C.pov.value] = "True"

        self.infos = []

        super().__init__(CharacterSearchLabels)

    def name(self):
        return self._data[C.name.value]

    def setName(self, value):
        self._data[C.name.value] = value

    def importance(self):
        return self._data[C.importance.value]

    def ID(self):
        return self._data[C.ID.value]

    def index(self, column=0):
        return self._model.indexFromItem(self, column)

    def data(self, column):
        if column == "Info":
            return self.infos
        else:
            return self._data.get(column, None)

    def assignRandomColor(self):
        """
        Assigns a random color the the character.
        """
        color = randomColor(QColor(Qt.white))
        self.setColor(color)

    def setColor(self, color):
        """
        Sets the character's color
        @param color: QColor.
        """
        px = QPixmap(32, 32)
        px.fill(color)
        self.icon = QIcon(px)
        try:
            self._model.dataChanged.emit(self.index(), self.index())
        except:
            # If it is the initialisation, won't be able to emit
            pass

    def color(self):
        """
        Returns character's color in QColor
        @return: QColor
        """
        return iconColor(self.icon)

    def setPOVEnabled(self, enabled):
        if enabled != self.pov():
            if enabled:
                self._data[C.pov.value] = 'True'
            else:
                self._data[C.pov.value] = 'False'

            try:
                self._model.dataChanged.emit(self.index(), self.index())
            except:
                # If it is the initialisation, won't be able to emit
                pass

    def pov(self):
        return self._data[C.pov.value] == 'True'

    def assignUniqueID(self, parent=QModelIndex()):
        """Assigns an unused character ID."""
        vals = []
        for c in self._model.characters:
            vals.append(int(c.ID()))

        k = 0
        while k in vals:
            k += 1

        self._data[C.ID.value] = str(k)

    def listInfos(self):
        r = []
        for i in self.infos:
            r.append((i.description, i.value))
        return r

    def searchTitle(self, column):
        return self.name()

    def searchOccurrences(self, searchRegex, column):
        results = []

        data = self.searchData(column)
        if isinstance(data, list):
            for i in range(0, len(data)):
                # For detailed info we will highlight the full row, so we pass the row index
                # to the highlighter instead of the (startPos, endPos) of the match itself.
                results += [self.wrapSearchOccurrence(column, i, 0, context) for
                            (startPos, endPos, context) in search(searchRegex, data[i].description)]
                results += [self.wrapSearchOccurrence(column, i, 0, context) for
                            (startPos, endPos, context) in search(searchRegex, data[i].value)]
        else:
            results += super().searchOccurrences(searchRegex, column)

        return results

    def searchID(self):
        return self.ID()

    def searchPath(self, column):
        return [self.translate("Characters"), self.name(), self.translate(self.searchColumnLabel(column))]

    def searchData(self, column):
        if column == C.infos:
            return self.infos
        else:
            return self.data(column)

    def searchModel(self):
        return Model.Character


class CharacterInfo():
    def __init__(self, character, description="", value=""):
        self.description = description
        self.value = value
        self.character = character



#!/usr/bin/env python
# --!-- coding: utf8 --!--


class abstractSearchResultHighlighter():
    """
    Interface for all classes highlighting search results on widgets.
    """
    def __init__(self):
        pass

    def highlightSearchResult(self, searchResult):
        raise NotImplementedError



#!/usr/bin/env python
# --!-- coding: utf8 --!--


from manuskript.models import references as Ref
from manuskript.functions import mainWindow
from manuskript.enums import  PlotStep
from PyQt5.QtWidgets import QTableView, QTextEdit
from manuskript.ui.highlighters.searchResultHighlighters.abstractSpecificSearchResultHighlighter import abstractSearchResultHighlighter


class plotStepSearchResultHighlighter(abstractSearchResultHighlighter):
    def __init__(self):
        super().__init__()

    def openView(self, searchResult):
        r = Ref.plotReference(searchResult.id())
        Ref.open(r)
        mainWindow().tabPlot.setEnabled(True)

    def retrieveWidget(self, searchResult):
        textEditMap = {
            PlotStep.name: [(1, "lstSubPlots", QTableView)],
            PlotStep.meta: [(1, "lstSubPlots", QTableView)],
            PlotStep.summary: [(1, "lstSubPlots", QTableView), (1, "txtSubPlotSummary", QTextEdit)]
        }

        map = textEditMap[searchResult.column()]
        widgets = []
        for tabIndex, widgetName, widgetClass in map:
            mainWindow().tabPlot.setCurrentIndex(tabIndex)

            widgets.append(mainWindow().tabPlot.findChild(widgetClass, widgetName))

        return widgets



#!/usr/bin/python
# -*- coding: utf-8 -*-

import re

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QBrush, QTextCursor, QColor, QFont, QSyntaxHighlighter
from PyQt5.QtGui import QTextBlockFormat, QTextCharFormat

import manuskript.models.references as Ref
import manuskript.ui.style as S
from manuskript import settings
from manuskript import functions as F

import logging
LOGGER = logging.getLogger(__name__)

class BasicHighlighter(QSyntaxHighlighter):
    def __init__(self, editor):
        QSyntaxHighlighter.__init__(self, editor.document())

        self.editor = editor
        self._defaultBlockFormat = QTextBlockFormat()
        self._defaultCharFormat = QTextCharFormat()
        self.defaultTextColor = QColor(S.text)
        self.backgroundColor = QColor(S.base)
        self.markupColor = QColor(S.textLight)
        self.linkColor = QColor(S.link)
        self.spellingErrorColor = QColor(Qt.red)

        # Matches during checking can be separated by their type (all of them listed here):
        # https://languagetool.org/development/api/org/languagetool/rules/ITSIssueType.html
        #
        # These are the colors for actual spell-, grammar- and style-checking:
        self._errorColors = {
            'addition' : QColor(255, 215, 0),               # gold
            'characters' : QColor(135, 206, 235),           # sky blue
            'duplication' : QColor(0, 255, 255),            # cyan / aqua
            'formatting' : QColor(0, 128, 128),             # teal
            'grammar' : QColor(0, 0, 255),                  # blue
            'inconsistency' : QColor(128, 128, 0),          # olive
            'inconsistententities' : QColor(46, 139, 87),   # sea green
            'internationalization' : QColor(255, 165, 0),   # orange
            'legal' : QColor(255, 69, 0),                   # orange red
            'length' : QColor(47, 79, 79),                  # dark slate gray
            'localespecificcontent' : QColor(188, 143, 143),# rosy brown
            'localeviolation' : QColor(128, 0, 0),          # maroon
            'markup' : QColor(128, 0, 128),                 # purple
            'misspelling' : QColor(255, 0, 0),              # red
            'mistranslation' : QColor(255, 0, 255),         # magenta / fuchsia
            'nonconformance' : QColor(255, 218, 185),       # peach puff
            'numbers' : QColor(65, 105, 225),               # royal blue
            'omission' : QColor(255, 20, 147),              # deep pink
            'other' : QColor(138, 43, 226),                 # blue violet
            'patternproblem' : QColor(0, 128, 0),           # green
            'register' : QColor(112,128,144),               # slate gray
            'style' : QColor(0, 255, 0),                    # lime
            'terminology' : QColor(0, 0, 128),              # navy
            'typographical' : QColor(255, 255, 0),          # yellow
            'uncategorized' : QColor(128, 128, 128),        # gray
            'untranslated' : QColor(210, 105, 30),          # chocolate
            'whitespace' : QColor(192, 192, 192)            # silver
        }

    def setDefaultBlockFormat(self, bf):
        self._defaultBlockFormat = bf
        self.rehighlight()

    def setDefaultCharFormat(self, cf):
        self._defaultCharFormat = cf
        self.rehighlight()

    def setMisspelledColor(self, color):
        self._errorColors['misspelled'] = color

    def updateColorScheme(self, rehighlight=True):
        """
        Generates a base set of colors that will take account of user
        preferences, and use system style.
        """

        # Reading user settings
        opt = settings.textEditor

        if not self.editor._fromTheme or not self.editor._themeData:

            self.defaultTextColor = QColor(opt["fontColor"])
            self.backgroundColor = (QColor(opt["background"])
                                    if not opt["backgroundTransparent"]
                                    else QColor(S.window))
            self.markupColor = F.mixColors(self.defaultTextColor,
                                           self.backgroundColor,
                                           .3)
            self.linkColor = QColor(S.link)
            self.spellingErrorColor = QColor(opt["misspelled"])
            self._defaultCharFormat.setForeground(QBrush(self.defaultTextColor))

        # FullscreenEditor probably
        else:
            opt = self.editor._themeData
            self.defaultTextColor = QColor(opt["Text/Color"])
            self.backgroundColor =  F.mixColors(
                QColor(opt["Foreground/Color"]),
                QColor(opt["Background/Color"]),
                int(opt["Foreground/Opacity"])/100.)
            self.markupColor = F.mixColors(self.defaultTextColor,
                                           self.backgroundColor,
                                           .3)
            self.linkColor = QColor(S.link)
            self.spellingErrorColor = QColor(opt["Text/Misspelled"])

        if rehighlight:
            self.rehighlight()

    def highlightBlock(self, text):
        """Apply syntax highlighting to the given block of text.
        """
        self.highlightBlockBefore(text)
        self.doHighlightBlock(text)
        self.highlightBlockAfter(text)

    def doHighlightBlock(self, text):
        """
        Virtual function to subclass.
        """
        pass

    def highlightBlockBefore(self, text):
        """Highlighting to do before anything else.

        When subclassing BasicHighlighter, you must call highlightBlockBefore
        before you do any custom highlighting. Or implement doHighlightBlock.
        """

        #LOGGER.debug("undoSteps before: %s", self.currentBlock().document().availableUndoSteps())
        c = QTextCursor(self.currentBlock())
        #c.joinPreviousEditBlock()
        bf = QTextBlockFormat(self._defaultBlockFormat)
        if bf != c.blockFormat():
            c.setBlockFormat(bf)
        #c.endEditBlock()
        #LOGGER.debug("undoSteps after: %s", self.currentBlock().document().availableUndoSteps())

        # self.setFormat(0, len(text), self._defaultCharFormat)

    def highlightBlockAfter(self, text):
        """Highlighting to do after everything else.

        When subclassing BasicHighlighter, you must call highlightBlockAfter
        after your custom highlighting. Or implement doHighlightBlock.
        """

        # References
        for txt in re.finditer(Ref.RegEx, text):
            fmt = self.format(txt.start())
            fmt.setFontFixedPitch(True)
            fmt.setFontWeight(QFont.DemiBold)

            if txt.group(1) == Ref.TextLetter:
                fmt.setBackground(QBrush(Ref.TextHighlightColor))
            elif txt.group(1) == Ref.CharacterLetter:
                fmt.setBackground(QBrush(Ref.CharacterHighlightColor))
            elif txt.group(1) == Ref.PlotLetter:
                fmt.setBackground(QBrush(Ref.PlotHighlightColor))
            elif txt.group(1) == Ref.WorldLetter:
                fmt.setBackground(QBrush(Ref.WorldHighlightColor))

            self.setFormat(txt.start(),
                           txt.end() - txt.start(),
                           fmt)

        if hasattr(self.editor, "spellcheck") and self.editor.spellcheck and self.editor._dict:
            # Spell checking

            # Following algorithm would not check words at the end of line.
            # This hacks adds a space to every line where the text cursor is not
            # So that it doesn't spellcheck while typing, but still spellchecks at
            # end of lines. See github's issue #166.
            textedText = text
            if self.currentBlock().position() + len(text) != \
               self.editor.textCursor().position():
                textedText = text + " "

            # The text should only be checked once as a whole
            for match in self.editor._dict.checkText(textedText):
                if match.locqualityissuetype in self._errorColors:
                    highlight_color = self._errorColors[match.locqualityissuetype]

                    format = self.format(match.start)
                    format.setUnderlineColor(highlight_color)
                    # SpellCheckUnderline fails with some fonts
                    format.setUnderlineStyle(QTextCharFormat.WaveUnderline)
                    self.setFormat(match.start, match.end - match.start, format)



#!/usr/bin/env python
# --!-- coding: utf8 --!--

from manuskript.models.searchResultModel import searchResultModel
from manuskript.enums import Character


def test_searchResultModel_constructionOk():
    searchResult = searchResultModel("Character", "3", Character.notes, "Lucas", "A > B > C", (15, 18), "This is <b>Lucas</b>")
    assert searchResult.id() == "3"
    assert searchResult.column() == Character.notes
    assert searchResult.title() == "Lucas"
    assert searchResult.path() == "A > B > C"
    assert searchResult.pos() == (15, 18)
    assert searchResult.context() == "This is <b>Lucas</b>"




#!/usr/bin/env python
# --!-- coding: utf8 --!--

from PyQt5.QtGui import QTextCursor
from PyQt5.QtWidgets import QTextEdit, QTableView, QListView, QLineEdit, QPlainTextEdit, QLabel


class widgetSelectionHighlighter():
    """
    Utility class for highlighting a search result on a widget.
    """
    def __init__(self):
        pass

    def highlight_widget_selection(self, widget, startPos, endPos, clearOnFocusOut=True):
        if isinstance(widget, QTextEdit) or isinstance(widget, QPlainTextEdit):
            self._highlightTextEditSearchResult(widget, startPos, endPos, clearOnFocusOut)
        elif isinstance(widget, QLineEdit):
            self._highlightLineEditSearchResult(widget, startPos, endPos, clearOnFocusOut)
        elif isinstance(widget, QTableView):
            self._highlightTableViewSearchResult(widget, startPos, clearOnFocusOut)
        elif isinstance(widget, QListView):
            self._highlightListViewSearchResult(widget, startPos, clearOnFocusOut)
        elif isinstance(widget, QLabel):
            self._highlightLabelSearchResult(widget, clearOnFocusOut)
        else:
            raise NotImplementedError

        widget.setFocus(True)

    @staticmethod
    def generateClearHandler(widget, clearCallback):
        """
        Generates a clear handler to be run when the given widget loses focus.

        :param widget: widget we want to attach the handler to
        :param clearCallback: callback to be called when the given widget loses focus.
        :return:
        """
        def clearHandler(_widget, previous_on_focus_out_event):
            clearCallback(_widget)
            _widget.focusOutEvent = previous_on_focus_out_event

        widget.focusOutEvent = lambda e: clearHandler(widget, widget.focusOutEvent)

    def _highlightTextEditSearchResult(self, textEdit, startPos, endPos, clearOnFocusOut):
        # On focus out, clear text edit selection.
        oldTextCursor = textEdit.textCursor()
        if clearOnFocusOut:
            self.generateClearHandler(textEdit, lambda widget: widget.setTextCursor(oldTextCursor))

        # Highlight search result on the text edit.
        c = textEdit.textCursor()
        c.setPosition(startPos)
        c.setPosition(endPos, QTextCursor.KeepAnchor)
        textEdit.setTextCursor(c)

    def _highlightLineEditSearchResult(self, lineEdit, startPos, endPos, clearOnFocusOut):
        # On focus out, clear line edit selection.
        if clearOnFocusOut:
            self.generateClearHandler(lineEdit, lambda widget: widget.deselect())

        # Highlight search result on line edit.
        lineEdit.setCursorPosition(startPos)
        lineEdit.cursorForward(True, endPos - startPos)

    def _highlightTableViewSearchResult(self, tableView, startPos, clearOnFocusOut):
        # On focus out, clear table selection.
        if clearOnFocusOut:
            self.generateClearHandler(tableView, lambda widget: widget.clearSelection())

        # Highlight table row containing search result.
        tableView.selectRow(startPos)

    def _highlightListViewSearchResult(self, listView, startPos, clearOnFocusOut):
        # On focus out, clear table selection.
        if clearOnFocusOut:
            self.generateClearHandler(listView, lambda widget: widget.selectionModel().clearSelection())

        # Highlight list item containing search result.
        listView.setCurrentIndex(listView.model().index(startPos, 0, listView.rootIndex()))

    def _highlightLabelSearchResult(self, label, clearOnFocusOut):
        # On focus out, clear label selection.
        # FIXME: This would overwrite all styles!
        oldStyle = label.styleSheet()
        if clearOnFocusOut:
            self.generateClearHandler(label, lambda widget: widget.setStyleSheet(oldStyle))

        # Highlight search result on label.
        label.setStyleSheet("background-color: steelblue")



#!/usr/bin/env python
# --!-- coding: utf8 --!--

from manuskript.enums import FlatData, Model
from manuskript.searchLabels import FlatDataSearchLabels

from manuskript.models.searchableModel import searchableModel
from manuskript.models.searchableItem import searchableItem

"""
All searches are performed on models inheriting from searchableModel, but special metadata such as book summaries
are stored directly on a GUI element (QStandardItemModel). We wrap this GUI element inside this wrapper class
so it exposes the same interface for searches.
"""
class flatDataModelWrapper(searchableModel, searchableItem):
    def __init__(self, qstandardItemModel):
        self.qstandardItemModel = qstandardItemModel

    def searchableItems(self):
        return [flatDataItemWrapper(self.qstandardItemModel)]


class flatDataItemWrapper(searchableItem):
    def __init__(self, qstandardItemModel):
        super().__init__(FlatDataSearchLabels)
        self.qstandardItemModel = qstandardItemModel

    def searchModel(self):
        return Model.FlatData

    def searchID(self):
        return None

    def searchTitle(self, column):
        return self.translate(self.searchColumnLabel(column))

    def searchPath(self, column):
        return [self.translate("Summary"), self.translate(self.searchColumnLabel(column))]

    def searchData(self, column):
        return self.qstandardItemModel.item(1, self.searchDataIndex(column)).text()

    @staticmethod
    def searchDataIndex(column):
        columnIndices = {
            FlatData.summarySituation: 0,
            FlatData.summarySentence: 1,
            FlatData.summaryPara: 2,
            FlatData.summaryPage: 3,
            FlatData.summaryFull: 4
        }

        return columnIndices[column]


#!/usr/bin/env python
# --!-- coding: utf8 --!--

from manuskript.enums import Outline, Character, FlatData, World, Plot, PlotStep

OutlineSearchLabels = {
    Outline.title: "Title",
    Outline.text: "Text",
    Outline.summarySentence: "One sentence summary",
    Outline.summaryFull: "Summary",
    Outline.POV: "POV",
    Outline.notes: "Notes",
    Outline.status: "Status",
    Outline.label: "Label"
}

CharacterSearchLabels = {
    Character.name: "Name",
    Character.motivation: "Motivation",
    Character.goal: "Goal",
    Character.conflict: "Conflict",
    Character.epiphany: "Epiphany",
    Character.summarySentence: "One sentence summary",
    Character.summaryPara: "One paragraph summary",
    Character.summaryFull: "Summary",
    Character.notes: "Notes",
    Character.infos: "Detailed info"
}

FlatDataSearchLabels = {
    FlatData.summarySituation: "Situation",
    FlatData.summarySentence: "One sentence summary",
    FlatData.summaryPara: "One paragraph summary",
    FlatData.summaryPage: "One page summary",
    FlatData.summaryFull: "Full summary"
}

WorldSearchLabels = {
    World.name: "Name",
    World.description: "Description",
    World.passion: "Passion",
    World.conflict: "Conflict"
}

# Search menu includes one single option for both plot and plotStep models. For plotStep related fields
# (like PlotStep.meta) we add an offset so it is not confused with the Plot enum value mapping to the same integer.
PLOT_STEP_COLUMNS_OFFSET = 30

PlotSearchLabels = {
    Plot.name: "Name",
    Plot.description: "Description",
    Plot.characters: "Characters",
    Plot.result: "Result",
    Plot.summary: "Summary",
    PLOT_STEP_COLUMNS_OFFSET + PlotStep.meta: "Meta"
}
